@@ -1,6 +1,6 @@
 // SPDX-License-Identifier: GPL-2.0
 // Copyright (c) 2020 Wenbo Zhang
-#include "vmlinux.h"
+#include <vmlinux.h>
 #include <bpf/bpf_helpers.h>
 #include <bpf/bpf_tracing.h>
 
@@ -9,14 +9,12 @@ struct {
 	__uint(max_entries, 10240);
 	__type(key, u32);
 	__type(value, u64);
-	__uint(map_flags, BPF_F_NO_PREALLOC);
 } start SEC(".maps");
 
 __u64 latency = 0;
 __u64 num = 0;
 
-SEC("fentry/migrate_misplaced_page")
-int BPF_PROG(migrate_misplaced_page)
+static int __migrate_misplaced_page(void)
 {
 	u32 pid = bpf_get_current_pid_tgid();
 	u64 ts = bpf_ktime_get_ns();
@@ -25,8 +23,19 @@ int BPF_PROG(migrate_misplaced_page)
 	return 0;
 }
 
-SEC("fexit/migrate_misplaced_page")
-int BPF_PROG(migrate_misplaced_page_exit)
+SEC("fentry/migrate_misplaced_page")
+int BPF_PROG(fentry_migrate_misplaced_page)
+{
+	return __migrate_misplaced_page();
+}
+
+SEC("kprobe/migrate_misplaced_page")
+int BPF_PROG(kprobe_migrate_misplaced_page)
+{
+	return __migrate_misplaced_page();
+}
+
+static int __migrate_misplaced_page_exit(void)
 {
 	u32 pid = bpf_get_current_pid_tgid();
 	u64 *tsp, ts = bpf_ktime_get_ns();
@@ -46,4 +55,16 @@ cleanup:
 	return 0;
 }
 
+SEC("fexit/migrate_misplaced_page")
+int BPF_PROG(fexit_migrate_misplaced_page_exit)
+{
+	return __migrate_misplaced_page_exit();
+}
+
+SEC("kretprobe/migrate_misplaced_page")
+int BPF_PROG(kretprobe_migrate_misplaced_page_exit)
+{
+	return __migrate_misplaced_page_exit();
+}
+
 char LICENSE[] SEC("license") = "GPL";
