@@ -5,6 +5,7 @@
 // Based on opensnoop(8) from BCC by Brendan Gregg and others.
 // 14-Feb-2020   Brendan Gregg   Created this.
 #include <argp.h>
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -15,7 +16,11 @@
 #include <bpf/bpf.h>
 #include "opensnoop.h"
 #include "opensnoop.skel.h"
+#include "btf_helpers.h"
 #include "trace_helpers.h"
+#ifdef USE_BLAZESYM
+#include "blazesym.h"
+#endif
 
 /* Tune the buffer size and wakeup rate. These settings cope with roughly
  * 50k opens/sec.
@@ -28,6 +33,12 @@
 
 #define NSEC_PER_SEC		1000000000ULL
 
+static volatile sig_atomic_t exiting = 0;
+
+#ifdef USE_BLAZESYM
+static blazesym *symbolizer;
+#endif
+
 static struct env {
 	pid_t pid;
 	pid_t tid;
@@ -39,6 +50,9 @@ static struct env {
 	bool extended;
 	bool failed;
 	char *name;
+#ifdef USE_BLAZESYM
+	bool callers;
+#endif
 } env = {
 	.uid = INVALID_UID
 };
@@ -50,7 +64,11 @@ const char argp_program_doc[] =
 "Trace open family syscalls\n"
 "\n"
 "USAGE: opensnoop [-h] [-T] [-U] [-x] [-p PID] [-t TID] [-u UID] [-d DURATION]\n"
+#ifdef USE_BLAZESYM
+"                 [-n NAME] [-e] [-c]\n"
+#else
 "                 [-n NAME] [-e]\n"
+#endif
 "\n"
 "EXAMPLES:\n"
 "    ./opensnoop           # trace all open() syscalls\n"
@@ -62,7 +80,11 @@ const char argp_program_doc[] =
 "    ./opensnoop -u 1000   # only trace UID 1000\n"
 "    ./opensnoop -d 10     # trace for 10 seconds only\n"
 "    ./opensnoop -n main   # only print process names containing \"main\"\n"
-"    ./opensnoop -e        # show extended fields\n";
+"    ./opensnoop -e        # show extended fields\n"
+#ifdef USE_BLAZESYM
+"    ./opensnoop -c        # show calling functions\n"
+#endif
+"";
 
 static const struct argp_option opts[] = {
 	{ "duration", 'd', "DURATION", 0, "Duration to trace"},
@@ -76,6 +98,9 @@ static const struct argp_option opts[] = {
 	{ "print-uid", 'U', NULL, 0, "Print UID"},
 	{ "verbose", 'v', NULL, 0, "Verbose debug output" },
 	{ "failed", 'x', NULL, 0, "Failed opens only"},
+#ifdef USE_BLAZESYM
+	{ "callers", 'c', NULL, 0, "Show calling functions"},
+#endif
 	{},
 };
 
@@ -143,6 +168,11 @@ static error_t parse_arg(int key, char *arg, struct argp_state *state)
 		}
 		env.uid = uid;
 		break;
+#ifdef USE_BLAZESYM
+	case 'c':
+		env.callers = true;
+		break;
+#endif
 	case ARGP_KEY_ARG:
 		if (pos_args++) {
 			fprintf(stderr,
@@ -157,47 +187,97 @@ static error_t parse_arg(int key, char *arg, struct argp_state *state)
 	return 0;
 }
 
-int libbpf_print_fn(enum libbpf_print_level level,
-		    const char *format, va_list args)
+static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
 {
 	if (level == LIBBPF_DEBUG && !env.verbose)
 		return 0;
 	return vfprintf(stderr, format, args);
 }
 
+static void sig_int(int signo)
+{
+	exiting = 1;
+}
+
 void handle_event(void *ctx, int cpu, void *data, __u32 data_sz)
 {
-	const struct event *e = data;
+	struct event e;
 	struct tm *tm;
+#ifdef USE_BLAZESYM
+	sym_src_cfg cfgs[] = {
+		{ .src_type = SRC_T_PROCESS, .params = { .process = { .pid = e->pid }}},
+	};
+	const blazesym_result *result = NULL;
+	const blazesym_csym *sym;
+	int i, j;
+#endif
+	int sps_cnt;
 	char ts[32];
 	time_t t;
 	int fd, err;
 
+	if (data_sz < sizeof(e)) {
+   	   	printf("Error: packet too small\n");
+   	   	return;
+	}
+	/* Copy data as alignment in the perf buffer isn't guaranteed. */
+	memcpy(&e, data, sizeof(e));
+
 	/* name filtering is currently done in user space */
-	if (env.name && strstr(e->comm, env.name) == NULL)
+	if (env.name && strstr(e.comm, env.name) == NULL)
 		return;
 
 	/* prepare fields */
 	time(&t);
 	tm = localtime(&t);
 	strftime(ts, sizeof(ts), "%H:%M:%S", tm);
-	if (e->ret >= 0) {
-		fd = e->ret;
+	if (e.ret >= 0) {
+		fd = e.ret;
 		err = 0;
 	} else {
 		fd = -1;
-		err = - e->ret;
+		err = - e.ret;
 	}
 
+#ifdef USE_BLAZESYM
+	if (env.callers)
+		result = blazesym_symbolize(symbolizer, cfgs, 1, (const uint64_t *)&e.callers, 2);
+#endif
+
 	/* print output */
-	if (env.timestamp)
+	sps_cnt = 0;
+	if (env.timestamp) {
 		printf("%-8s ", ts);
-	if (env.print_uid)
-		printf("%-6d ", e->uid);
-	printf("%-6d %-16s %3d %3d ", e->pid, e->comm, fd, err);
-	if (env.extended)
-		printf("%08o ", e->flags);
-	printf("%s\n", e->fname);
+		sps_cnt += 9;
+	}
+	if (env.print_uid) {
+		printf("%-7d ", e.uid);
+		sps_cnt += 8;
+	}
+	printf("%-6d %-16s %3d %3d ", e.pid, e.comm, fd, err);
+	sps_cnt += 7 + 17 + 4 + 4;
+	if (env.extended) {
+		printf("%08o ", e.flags);
+		sps_cnt += 9;
+	}
+	printf("%s\n", e.fname);
+
+#ifdef USE_BLAZESYM
+	for (i = 0; result && i < result->size; i++) {
+		if (result->entries[i].size == 0)
+			continue;
+		sym = &result->entries[i].syms[0];
+
+		for (j = 0; j < sps_cnt; j++)
+			printf(" ");
+		if (sym->line_no)
+			printf("%s:%ld\n", sym->symbol, sym->line_no);
+		else
+			printf("%s\n", sym->symbol);
+	}
+
+	blazesym_result_free(result);
+#endif
 }
 
 void handle_lost_events(void *ctx, int cpu, __u64 lost_cnt)
@@ -207,12 +287,12 @@ void handle_lost_events(void *ctx, int cpu, __u64 lost_cnt)
 
 int main(int argc, char **argv)
 {
+	LIBBPF_OPTS(bpf_object_open_opts, open_opts);
 	static const struct argp argp = {
 		.options = opts,
 		.parser = parse_arg,
 		.doc = argp_program_doc,
 	};
-	struct perf_buffer_opts pb_opts;
 	struct perf_buffer *pb = NULL;
 	struct opensnoop_bpf *obj;
 	__u64 time_end = 0;
@@ -224,13 +304,13 @@ int main(int argc, char **argv)
 
 	libbpf_set_print(libbpf_print_fn);
 
-	err = bump_memlock_rlimit();
+	err = ensure_core_btf(&open_opts);
 	if (err) {
-		fprintf(stderr, "failed to increase rlimit: %d\n", err);
+		fprintf(stderr, "failed to fetch necessary BTF for CO-RE: %s\n", strerror(-err));
 		return 1;
 	}
 
-	obj = opensnoop_bpf__open();
+	obj = opensnoop_bpf__open_opts(&open_opts);
 	if (!obj) {
 		fprintf(stderr, "failed to open BPF object\n");
 		return 1;
@@ -242,15 +322,11 @@ int main(int argc, char **argv)
 	obj->rodata->targ_uid = env.uid;
 	obj->rodata->targ_failed = env.failed;
 
-#ifdef __aarch64__
-	/* aarch64 has no open syscall, only openat variants.
-	 * Disable associated tracepoints that do not exist. See #3344.
-	 */
-	bpf_program__set_autoload(
-		obj->progs.tracepoint__syscalls__sys_enter_open, false);
-	bpf_program__set_autoload(
-		obj->progs.tracepoint__syscalls__sys_exit_open, false);
-#endif
+	/* aarch64 and riscv64 don't have open syscall */
+	if (!tracepoint_exists("syscalls", "sys_enter_open")) {
+		bpf_program__set_autoload(obj->progs.tracepoint__syscalls__sys_enter_open, false);
+		bpf_program__set_autoload(obj->progs.tracepoint__syscalls__sys_exit_open, false);
+	}
 
 	err = opensnoop_bpf__load(obj);
 	if (err) {
@@ -264,24 +340,31 @@ int main(int argc, char **argv)
 		goto cleanup;
 	}
 
+#ifdef USE_BLAZESYM
+	if (env.callers)
+		symbolizer = blazesym_new();
+#endif
+
 	/* print headers */
 	if (env.timestamp)
 		printf("%-8s ", "TIME");
 	if (env.print_uid)
-		printf("%-6s ", "UID");
+		printf("%-7s ", "UID");
 	printf("%-6s %-16s %3s %3s ", "PID", "COMM", "FD", "ERR");
 	if (env.extended)
 		printf("%-8s ", "FLAGS");
-	printf("%s\n", "PATH");
+	printf("%s", "PATH");
+#ifdef USE_BLAZESYM
+	if (env.callers)
+		printf("/CALLER");
+#endif
+	printf("\n");
 
 	/* setup event callbacks */
-	pb_opts.sample_cb = handle_event;
-	pb_opts.lost_cb = handle_lost_events;
 	pb = perf_buffer__new(bpf_map__fd(obj->maps.events), PERF_BUFFER_PAGES,
-			      &pb_opts);
-	err = libbpf_get_error(pb);
-	if (err) {
-		pb = NULL;
+			      handle_event, handle_lost_events, NULL, NULL);
+	if (!pb) {
+		err = -errno;
 		fprintf(stderr, "failed to open perf buffer: %d\n", err);
 		goto cleanup;
 	}
@@ -290,19 +373,32 @@ int main(int argc, char **argv)
 	if (env.duration)
 		time_end = get_ktime_ns() + env.duration * NSEC_PER_SEC;
 
+	if (signal(SIGINT, sig_int) == SIG_ERR) {
+		fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
+		err = 1;
+		goto cleanup;
+	}
+
 	/* main: poll */
-	while (1) {
-		usleep(PERF_BUFFER_TIME_MS * 1000);
-		if ((err = perf_buffer__poll(pb, PERF_POLL_TIMEOUT_MS)) < 0)
-			break;
+	while (!exiting) {
+		err = perf_buffer__poll(pb, PERF_POLL_TIMEOUT_MS);
+		if (err < 0 && err != -EINTR) {
+			fprintf(stderr, "error polling perf buffer: %s\n", strerror(-err));
+			goto cleanup;
+		}
 		if (env.duration && get_ktime_ns() > time_end)
 			goto cleanup;
+		/* reset err to return 0 if exiting */
+		err = 0;
 	}
-	printf("Error polling perf buffer: %d\n", err);
 
 cleanup:
 	perf_buffer__free(pb);
 	opensnoop_bpf__destroy(obj);
+	cleanup_core_btf(&open_opts);
+#ifdef USE_BLAZESYM
+	blazesym_free(symbolizer);
+#endif
 
 	return err != 0;
 }
