@@ -44,6 +44,8 @@ Location::Location(uint64_t addr, const std::string &bin_path, const char *arg_f
   ArgumentParser_powerpc64 parser(arg_fmt);
 #elif __s390x__
   ArgumentParser_s390x parser(arg_fmt);
+#elif __riscv
+  ArgumentParser_riscv64 parser(arg_fmt);
 #else
   ArgumentParser_x64 parser(arg_fmt);
 #endif
