@@ -8,14 +8,14 @@ from bcc import BPF
 
 class TestSharedTable(unittest.TestCase):
     def test_close_extern(self):
-        b1 = BPF(text="""BPF_TABLE_PUBLIC("array", int, int, table1, 10);""")
+        b1 = BPF(text=b"""BPF_TABLE_PUBLIC("array", int, int, table1, 10);""")
 
-        with BPF(text="""BPF_TABLE("extern", int, int, table1, 10);""") as b2:
-            t2 = b2["table1"]
+        with BPF(text=b"""BPF_TABLE("extern", int, int, table1, 10);""") as b2:
+            t2 = b2[b"table1"]
             t2[ct.c_int(1)] = ct.c_int(10)
             self.assertEqual(len(t2), 10)
 
-        t1 = b1["table1"]
+        t1 = b1[b"table1"]
         self.assertEqual(t1[ct.c_int(1)].value, 10)
         self.assertEqual(len(t1), 10)
 
